"use strict";
/**
 * @fileOverview spread sheet converter
 * @name index
 * @author Yuhei Aihara <yu.e.yu.4119@gmail.com>
 * https://github.com/iyu/spread2json
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = __importDefault(require("lodash"));
const api_1 = __importDefault(require("./api"));
const logging_1 = __importDefault(require("./logging"));
/**
 * util
 */
/**
 * column key map 'A','B','C',,,'AA','AB',,,'AAA','AAB',,,'ZZZ'
 */
const COLUMN_KEYMAP = (() => {
    const map = [];
    /**
     * @param {number} digit
     * @param {string} value
     */
    function genValue(digit, value) {
        let _value;
        for (let i = 65; i < 91; i++) {
            _value = value + String.fromCharCode(i);
            if (digit > 1) {
                genValue(digit - 1, _value);
            }
            else {
                map.push(_value);
            }
        }
    }
    // A~Z
    genValue(1, '');
    // AA~ZZ
    genValue(2, '');
    // AAA~ZZZ
    genValue(3, '');
    return map;
})();
/**
 * column to number
 */
const columnToNumber = (column) => {
    return COLUMN_KEYMAP.indexOf(column) + 1;
};
/**
 * separate cell
 * @param {string} cell
 * @returns {Object}
 * @example
 * console.log(separateCell('A1'));
 * > { cell: 'A1', column: 'A', row: 1 }
 */
const separateCell = (cell) => {
    const match = cell.match(/^(\D+)(\d+)$/) || [];
    return {
        cell,
        column: match[1],
        col: columnToNumber(match[1]),
        row: parseInt(match[2], 10),
    };
};
const genWorksheetLink = lodash_1.default.template('https://docs.google.com/spreadsheets/d/<%= spreadsheetId %>/edit#gid=<%= sheetId %>');
class Spread2Json {
    constructor() {
        this.opts = {
            // Cell with a custom sheet option.
            option_cell: 'A1',
            // Line with a data attribute.
            attr_line: 2,
            // Line with a attribute description.
            desc_line: 3,
            // Line with a data.
            data_line: 4,
            // ref key
            ref_keys: ['_id'],
            // Custom logger.
            logger: undefined,
            // Google API options.
            api: {
                client_id: undefined,
                client_secret: undefined,
                redirect_url: 'http://localhost',
                token_file: {
                    use: true,
                    path: './dist/token.json',
                },
            },
        };
        this._parser = {
            string: (d) => {
                return String(d);
            },
            number: (d) => {
                return Number(d);
            },
            num: (d) => {
                return this._parser.number(d);
            },
            boolean: (d) => {
                if (typeof d === 'boolean') {
                    return d;
                }
                if (typeof d === 'string') {
                    return d.toLowerCase() !== 'false' && d !== '0';
                }
                return !!d;
            },
            bool: (d) => {
                return this._parser.boolean(d);
            },
            date: (d) => {
                return new Date(d).getTime();
            },
            auto: (d) => {
                return Number.isNaN(Number(d)) ? d : this._parser.number(d);
            },
        };
        this._stringify = {
            string: (d) => {
                return d;
            },
            number: (d) => {
                return d;
            },
            num: (d) => {
                return this._stringify.number(d);
            },
            boolean: (d) => {
                return d;
            },
            bool: (d) => {
                return this._stringify.boolean(d);
            },
            date: (d) => {
                const date = new Date(d);
                const yyyy = date.getFullYear();
                const mm = `${`0${date.getMonth() + 1}`.slice(-2)}`;
                const dd = `${`0${date.getDate()}`.slice(-2)}`;
                const h = `${`0${date.getHours()}`.slice(-2)}`;
                const m = `${`0${date.getMinutes()}`.slice(-2)}`;
                const s = `${`0${date.getSeconds()}`.slice(-2)}`;
                return `${yyyy}/${mm}/${dd} ${h}:${m}:${s}`;
            },
            auto: (d) => {
                return d;
            },
        };
    }
    /**
     * API function
     */
    generateAuthUrl(opts) { return api_1.default.generateAuthUrl(opts); }
    async getAccessToken(code, callback) {
        if (!callback) {
            callback = (err, result) => {
                if (err) {
                    throw err;
                }
                return result;
            };
        }
        try {
            const result = await api_1.default.getAccessToken(code);
            return callback(null, result);
        }
        catch (e) {
            return callback(e);
        }
    }
    async refreshAccessToken(credentials, callback) {
        if (!callback) {
            callback = (err, result) => {
                if (err) {
                    throw err;
                }
                return result;
            };
        }
        try {
            const result = await api_1.default.refreshAccessToken(credentials);
            return callback(null, result);
        }
        catch (e) {
            return callback(e);
        }
    }
    /**
     * setup
     * @param {Object} options
     * @param {string} [options.option_cell='A1']
     * @param {number} [options.attr_line=2]
     * @param {number} [options.data_line=4]
     * @param {string} [options.ref_key]
     * @param {string[]} [options.ref_keys=['_id']]
     * @param {Object} options.api
     * @param {string} options.api.client_id
     * @param {string} options.api.client_secret
     * @param {string} [options.api.redirect_url='http://localhost']
     * @param {Object} [options.api.token_file]
     * @param {boolean} [options.api.token_file.use=true]
     * @param {string} [options.api.token_file.path='./dist/token.json']
     * @param {Logger} [logger]
     * @example
     * var options = {
     *     option_cell: 'A1',
     *     attr_line: 2,
     *     data_line: 4,
     *     ref_keys: ['_id'],
     *     api: {
     *         client_id: 'YOUR CLIENT ID HERE',
     *         client_secret: 'YOUR CLIENT SECRET HERE',
     *         redirect_url: 'http://localhost',
     *         token_file: {
     *           use: true
     *           path: './dist/token.json'
     *         }
     *     },
     *     logger: CustomLogger
     * }
     */
    setup(options) {
        const opts = lodash_1.default.assign(this.opts, options);
        if (opts.ref_key) {
            opts.ref_keys = options.ref_keys || [opts.ref_key];
            delete opts.ref_key;
        }
        if (opts.logger) {
            logging_1.default.logger = opts.logger;
        }
        api_1.default.setup.call(api_1.default, opts.api);
        logging_1.default.logger.info('spread2json setup');
    }
    async getSpreadsheet(credentials, opts, callback) {
        if (arguments.length === 1) {
            if (typeof credentials === 'function') {
                callback = credentials;
                opts = undefined;
                credentials = undefined;
            }
            else if (!lodash_1.default.has(credentials, ['access_token'])) {
                opts = credentials;
                credentials = undefined;
            }
        }
        else if (arguments.length === 2) {
            if (typeof opts === 'function') {
                callback = opts;
                if (!lodash_1.default.has(credentials, ['access_token'])) {
                    opts = credentials;
                    credentials = undefined;
                }
                else {
                    opts = undefined;
                }
            }
        }
        if (!callback) {
            callback = (err, result) => {
                if (err) {
                    throw err;
                }
                return result;
            };
        }
        let result;
        try {
            result = await api_1.default.getSpreadsheet(credentials, opts);
        }
        catch (e) {
            return callback(e);
        }
        const list = lodash_1.default.map(result, (sheet) => {
            const data = lodash_1.default.assign({}, sheet, {
                link: genWorksheetLink({
                    spreadsheetId: sheet.id,
                    sheetId: 0,
                }),
            });
            return data;
        });
        return callback(null, list);
    }
    async getWorksheet(credentials, spreadsheetId, callback) {
        if (!lodash_1.default.has(credentials, 'access_token')) {
            callback = spreadsheetId;
            spreadsheetId = credentials;
            credentials = undefined;
        }
        if (!callback) {
            callback = (err, result) => {
                if (err) {
                    throw err;
                }
                return result;
            };
        }
        let result;
        try {
            result = await api_1.default.getWorksheet(credentials, spreadsheetId);
        }
        catch (e) {
            return callback(e);
        }
        const list = lodash_1.default.map(lodash_1.default.get(result, 'sheets'), (sheet) => {
            const sheetId = lodash_1.default.get(sheet, ['properties', 'sheetId'], 0);
            const data = {
                title: lodash_1.default.get(sheet, ['properties', 'title'], ''),
                sheetId,
                link: genWorksheetLink({
                    spreadsheetId,
                    sheetId,
                }),
            };
            return data;
        });
        return callback(null, list);
    }
    async addWorksheet(credentials, spreadsheetId, title, rowCount, colCount, callback) {
        if (!lodash_1.default.has(credentials, ['access_token'])) {
            callback = colCount;
            colCount = rowCount;
            rowCount = title;
            title = spreadsheetId;
            spreadsheetId = credentials;
            credentials = undefined;
        }
        rowCount = lodash_1.default.isNumber(rowCount) ? rowCount : 50;
        colCount = lodash_1.default.isNumber(colCount) ? colCount : 10;
        if (!callback) {
            callback = (err, result) => {
                if (err) {
                    throw err;
                }
                return result;
            };
        }
        let result;
        try {
            result = await api_1.default.addWorksheet(credentials, spreadsheetId, title, rowCount, colCount);
        }
        catch (e) {
            return callback(e);
        }
        const sheetId = lodash_1.default.get(result, ['replies', 0, 'addSheet', 'properties', 'sheetId'], 0);
        const data = {
            title: lodash_1.default.get(result, ['replies', 0, 'addSheet', 'properties', 'title'], ''),
            sheetId,
            link: genWorksheetLink({
                spreadsheetId,
                sheetId,
            }),
        };
        return callback(null, data);
    }
    /**
     * format
     * @param {Object[]} cells
     * @param {string} cells[].cell
     * @param {string} cells[].column
     * @param {number} cells[].row
     * @param {string} cells[].value
     * @private
     * @example
     * var cells = [
     *     { cell: 'A1', column: 'A', row: 1, value: '{}' },,,,
     * ]
     */
    _format(cells) {
        const opts = {
            attr_line: this.opts.attr_line,
            data_line: this.opts.data_line,
            ref_keys: this.opts.ref_keys,
            format: {},
        };
        let beforeRow;
        let idx = {};
        const list = [];
        lodash_1.default.forEach(cells, (cell) => {
            if (beforeRow !== cell.row) {
                lodash_1.default.forEach(idx, (i) => {
                    if (i.type !== 'format') {
                        i.value += 1;
                    }
                });
                beforeRow = cell.row;
            }
            if (cell.cell === this.opts.option_cell) {
                const _opts = JSON.parse(cell.value);
                if (_opts.ref_key && !_opts.ref_keys) {
                    _opts.ref_keys = [_opts.ref_key];
                    delete _opts.ref_key;
                }
                lodash_1.default.assign(opts, _opts);
                return;
            }
            if (cell.row === opts.attr_line) {
                const type = cell.value.match(/:(\w+)$/);
                const key = cell.value.replace(/:\w+$/, '');
                const keys = key.split('.');
                opts.format[cell.column] = {
                    type: (type && type[1] && type[1].toLowerCase()) || '',
                    key,
                    keys,
                };
                return;
            }
            const format = opts.format[cell.column];
            if (cell.row < opts.data_line || !format) {
                return;
            }
            if (format.type === 'index') {
                const _idx = parseInt(cell.value, 10);
                if (!idx[format.key] || idx[format.key].value !== _idx) {
                    idx[format.key] = {
                        type: 'format',
                        value: _idx,
                    };
                    lodash_1.default.forEach(idx, (i, key) => {
                        if (new RegExp(`^${format.key}.+$`).test(key)) {
                            idx[key].value = 0;
                        }
                    });
                }
                return;
            }
            if (((!opts.type || opts.type === 'origin') && format.key === opts.ref_keys[0]) ||
                format.key === '__ref' || format.key === '__ref_0') {
                idx = {};
                list.push({});
            }
            let data = lodash_1.default.last(list);
            lodash_1.default.forEach(format.keys, (_key, i) => {
                const isArray = /^#/.test(_key);
                const isSplitArray = /^\$/.test(_key);
                let __key;
                if (isArray) {
                    _key = _key.replace(/^#/, '');
                    data[_key] = data[_key] || [];
                }
                if (isSplitArray) {
                    _key = _key.replace(/^\$/, '');
                }
                if (i + 1 !== format.keys.length) {
                    if (isArray) {
                        __key = format.keys.slice(0, i + 1).join('.');
                        let _idx = idx[__key];
                        if (!_idx) {
                            idx[__key] = {
                                type: 'normal',
                                value: data[_key].length ? data[_key].length - 1 : 0,
                            };
                            _idx = idx[__key];
                        }
                        data[_key][_idx.value] = data[_key][_idx.value] || {};
                        data = data[_key][_idx.value];
                        return;
                    }
                    data[_key] = data[_key] || {};
                    data = data[_key];
                    return;
                }
                if (isArray) {
                    __key = format.keys.slice(0, i + 1).join('.');
                    let _idx = idx[__key];
                    if (!_idx) {
                        idx[__key] = {
                            type: 'normal',
                            value: data[_key].length ? data[_key].length - 1 : 0,
                        };
                        _idx = idx[__key];
                    }
                    data = data[_key];
                    _key = String(_idx.value);
                }
                if (data[_key]) {
                    return;
                }
                const type = format.type;
                if (this._parser[type]) {
                    data[_key] = isSplitArray ? cell.value.split(',').map(this._parser[type]) : this._parser[type](cell.value);
                }
                else {
                    data[_key] = isSplitArray ? cell.value.split(',') : cell.value;
                }
            });
        });
        return {
            opts,
            list,
        };
    }
    /**
     * find origin data
     * @param {Object} dataMap
     * @param {Object} opts
     * @param {Object} data
     * @private
     */
    _findOrigin(dataMap, opts, data) {
        let origin;
        if (data.__ref) {
            origin = dataMap[data.__ref];
        }
        else {
            let dataKeys = '';
            lodash_1.default.forEach(opts.ref_keys, (refKey, i) => {
                dataKeys += data[`__ref_${i}`];
                if (i < opts.ref_keys.length - 1) {
                    dataKeys += '.';
                }
            });
            origin = dataMap[dataKeys];
        }
        if (!origin || !opts.key) {
            logging_1.default.logger.error('not found origin.', JSON.stringify(data));
            return;
        }
        const keys = opts.key.split('.');
        const __in = data.__in ? data.__in.split('.') : [];
        for (let i = 0; i < keys.length; i++) {
            if (/^#/.test(keys[i])) {
                const key = keys[i].replace(/^#/, '');
                const index = __in[i] && __in[i].replace(/^#.+:(\d+)$/, '$1');
                if (!index) {
                    logging_1.default.logger.error('not found index.', JSON.stringify(data));
                    return;
                }
                origin[key] = origin[key] || [];
                origin = origin[key];
                origin[index] = origin[index] || {};
                origin = origin[index];
            }
            else if (keys[i] === '$') {
                origin = origin[__in[i]];
            }
            else if (i + 1 === keys.length) {
                origin[keys[i]] = origin[keys[i]] || (opts.type === 'array' ? [] : {});
                origin = origin[keys[i]];
            }
            else {
                origin[keys[i]] = origin[keys[i]] || {};
                origin = origin[keys[i]];
            }
            if (!origin) {
                logging_1.default.logger.error('not found origin parts.', JSON.stringify(data));
                return;
            }
        }
        if (opts.type === 'array') {
            if (!Array.isArray(origin)) {
                logging_1.default.logger.error('is not Array.', JSON.stringify(data));
                return;
            }
            origin.push({});
            origin = origin[origin.length - 1];
        }
        else if (opts.type === 'map') {
            if (!data.__key) {
                logging_1.default.logger.error('not found __key.', JSON.stringify(data));
                return;
            }
            origin[data.__key] = {};
            origin = origin[data.__key];
        }
        else {
            logging_1.default.logger.error(opts);
            return;
        }
        return origin;
    }
    async getWorksheetDatas(credentials, spreadsheetId, worksheetNames, callback) {
        if (!lodash_1.default.has(credentials, 'access_token')) {
            callback = worksheetNames;
            worksheetNames = spreadsheetId;
            spreadsheetId = credentials;
            credentials = undefined;
        }
        if (!callback) {
            callback =
                (err, result, errList) => {
                    if (err) {
                        throw err;
                    }
                    return [result, errList];
                };
        }
        let errList;
        let results;
        try {
            results = await Promise.all(lodash_1.default.map(worksheetNames, async (sheetName) => {
                const result = await api_1.default.getList(credentials, spreadsheetId, sheetName);
                const cells = lodash_1.default.transform(result.values, (ret, columns, i) => {
                    lodash_1.default.forEach(columns, (value, j) => {
                        if (value === '') {
                            return;
                        }
                        const rowName = i + 1;
                        const columnName = COLUMN_KEYMAP[j];
                        ret.push({
                            cell: columnName + rowName,
                            column: columnName,
                            row: rowName,
                            value,
                        });
                    });
                }, []);
                let _result;
                try {
                    _result = lodash_1.default.assign({ name: '' }, this._format(cells));
                }
                catch (e) {
                    logging_1.default.logger.error('invalid sheet format.', sheetName);
                    errList = errList || [];
                    errList.push({
                        name: sheetName,
                        error: e,
                    });
                    return;
                }
                _result.name = sheetName;
                return _result;
            }));
        }
        catch (e) {
            return callback(e);
        }
        return callback(null, lodash_1.default.compact(results), errList);
    }
    /**
     * sheetDatas to json
     * @param {Object[]} sheetDatas
     * @param {string} sheetDatas.name - work sheet name
     * @param {Object} sheetDatas.opts - sheet option
     * @param {Object[]} sheetDatas.list - sheet data list
     * @param {Function} [callback]
     */
    toJson(sheetDatas, callback) {
        const collectionMap = {};
        const optionMap = {};
        const errors = {};
        if (!callback) {
            callback = (err, _collectionMap, _optionMap) => {
                if (err) {
                    throw err;
                }
                return [_collectionMap, _optionMap];
            };
        }
        lodash_1.default.forEach(sheetDatas, (sheetData) => {
            const opts = sheetData.opts;
            const name = opts.name || sheetData.name;
            const refKeys = opts.ref_keys;
            collectionMap[name] = collectionMap[name] || {};
            const dataMap = collectionMap[name];
            if (!optionMap[name]) {
                optionMap[name] = opts;
            }
            else {
                optionMap[name] = lodash_1.default.assign({}, opts, optionMap[name]);
                lodash_1.default.assign(optionMap[name].format, opts.format);
            }
            lodash_1.default.forEach(sheetData.list, (data) => {
                if (!opts.type || opts.type === 'origin') {
                    let keys = '';
                    lodash_1.default.forEach(refKeys, (refKey, i) => {
                        keys += data[refKey];
                        if (i < refKeys.length - 1) {
                            keys += '.';
                        }
                    });
                    dataMap[keys] = data;
                }
                else {
                    const origin = this._findOrigin(dataMap, opts, data);
                    if (origin) {
                        lodash_1.default.forEach(data, (d, k) => {
                            if (/^__ref/.test(k)) {
                                delete data[k];
                            }
                        });
                        delete data.__in;
                        delete data.__key;
                        lodash_1.default.assign(origin, data);
                    }
                    else {
                        errors[name] = errors[name] || [];
                        errors[name].push(data);
                    }
                }
            });
        });
        return callback(lodash_1.default.isEmpty(errors) ? null : errors, collectionMap, optionMap);
    }
    /**
     * jsonschema to spread2json format
     * @param {Object} schema - jsonschema
     * @param {string[]} refkeys - ref keys
     * @param {string} [path=''] - data path
     */
    convertSchema2Format(schema, refKeys, path) {
        path = path || '';
        let result = [];
        switch (schema.type) {
            case 'string':
            case 'boolean':
                result.push(`${path}:${schema.type}`);
                break;
            case 'number':
            case 'integer':
                if (schema.format === 'date-time') {
                    result.push(`${path}:date`);
                }
                else {
                    result.push(`${path}:number`);
                }
                break;
            case 'object':
                lodash_1.default.forEach(schema.properties, (sc, key) => {
                    let nextKey = key;
                    if (sc.type === 'array' && sc.items) {
                        if (sc.items.type === 'object') {
                            nextKey = `#${key}`;
                        }
                        else if (sc.items.type !== 'array') {
                            nextKey = `$${key}`;
                        }
                    }
                    const nextPath = path ? `${path}.${nextKey}` : nextKey;
                    const format = this.convertSchema2Format(sc, refKeys, nextPath);
                    result.push.apply(result, format);
                });
                lodash_1.default.forEach(schema.allOf, (sc) => {
                    const format = this.convertSchema2Format(sc, refKeys, path);
                    result.push.apply(result, format);
                    result = lodash_1.default.uniq(result);
                });
                lodash_1.default.forEach(schema.oneOf, (sc) => {
                    const format = this.convertSchema2Format(sc, refKeys, path);
                    result.push.apply(result, format);
                    result = lodash_1.default.uniq(result);
                });
                break;
            case 'array': {
                if (schema.items) {
                    const arrayMark = schema.items.type === 'object' ? '#' : '$';
                    const format = this.convertSchema2Format(schema.items, refKeys, path || arrayMark);
                    result.push.apply(result, format);
                }
                else {
                    throw new Error(`invalid schema type. ${schema.type}`);
                }
                break;
            }
            default:
                throw new Error(`invalid schema type. ${schema.type}`);
        }
        return lodash_1.default.sortBy(result, (n, index) => {
            const i = lodash_1.default.indexOf(refKeys, n.replace(/:\w+$/, ''));
            return i === -1 ? index : (refKeys.length - i) * -1;
        });
    }
    /**
     * multiple sheet
     * @param {string} name
     * @param {string[]} refKeys
     * @param {string[]} format
     * @param {Object[]} datas
     */
    splitSheetData(name, refKeys, format, datas) {
        const map = {
            [name]: {
                name,
                opts: lodash_1.default.defaults({
                    name,
                    ref_keys: refKeys,
                }, {
                    attr_line: this.opts.attr_line,
                    desc_line: this.opts.desc_line,
                    data_line: this.opts.data_line,
                    ref_keys: this.opts.ref_keys,
                }),
                format: [],
                description: [],
                datas,
            },
        };
        const matches = [];
        const checks = [];
        lodash_1.default.forEach(format, (keys) => {
            const checked = lodash_1.default.some(checks, (regexp) => {
                return regexp.test(keys);
            });
            if (checked) {
                return;
            }
            let match = keys.match(/#/g);
            if (match && match.length >= 2) {
                // keys:'obj.#arr.obj2.#arr2.str' -> match:[keys,'obj.','arr']
                match = keys.match(/^([^#]+)?#([^#.]+)\..+$/);
                if (match) {
                    matches.push(match);
                    // ^obj.#arr\.
                    checks.push(new RegExp(`^${(match[1] ? match[1] : '')}#${match[2]}\\.`));
                }
            }
        });
        lodash_1.default.forEach(format, (keys) => {
            let match = [];
            let check = '';
            const checked = lodash_1.default.some(checks, (regexp, i) => {
                match = matches[i] || [];
                check = regexp;
                return regexp.test(keys);
            });
            if (!checked) {
                map[name].format.push(keys);
                map[name].description.push(lodash_1.default.last(keys.replace(/#|\$/g, '').replace(/:.+$/, '').split('.')) || '');
                return;
            }
            const key = (match[1] ? match[1] : '') + match[2];
            const _sheetName = `${name}.${key}`;
            const _keys = keys.replace(check, '');
            if (!map[_sheetName]) {
                map[_sheetName] = lodash_1.default.clone(map[name]);
                map[_sheetName].name = _sheetName;
                map[_sheetName].opts = lodash_1.default.assign({ type: 'array', key }, map[name].opts);
                map[_sheetName].format = lodash_1.default.map(refKeys, (refKey, i) => {
                    return `__ref_${i}`;
                });
                map[_sheetName].description = lodash_1.default.clone(refKeys);
            }
            map[_sheetName].format.push(_keys);
            map[_sheetName].description.push(lodash_1.default.last(_keys.replace(/#|\$/g, '.').replace(/:.+$/, '').split('.')) || '');
        });
        return lodash_1.default.values(map);
    }
    /**
     * sheetDatas to cells
     * FIXME array of array is unsupported
     * @param {Object} worksheetData
     * @param {Object} [worksheetData.opts={}] - sheet options
     * @param {string[]} worksheetData.format - data attribute
     * @param {string[]} worksheetData.description - attribute description
     * @param {Object[]} worksheetData.datas - datas
     * @param {string} [worksheetData.name] - sheet name
     */
    toCells(worksheetData) {
        const self = this;
        const cells = [];
        const opts = worksheetData.opts || {};
        const maxCol = worksheetData.format.length;
        const sheetName = worksheetData.name;
        let maxRow = 0;
        // opts
        const cr = separateCell(this.opts.option_cell);
        cells.push({
            col: cr.col,
            row: cr.row,
            value: JSON.stringify(opts),
        });
        // format
        lodash_1.default.forEach(worksheetData.format, (attr, i) => {
            cells.push({
                col: i + 1,
                row: this.opts.attr_line,
                value: attr,
            });
        });
        // description
        lodash_1.default.forEach(worksheetData.description, (desc, i) => {
            cells.push({
                col: i + 1,
                row: this.opts.desc_line,
                value: desc,
            });
        });
        // datas
        maxRow = this.opts.data_line;
        let row = maxRow;
        /**
         * add cell data to cells
         * @private
         */
        function addCell(data) {
            lodash_1.default.forEach(worksheetData.format, (attr, i) => {
                try {
                    const last = lodash_1.default.last(attr.split('.')) || '';
                    const hasArray = /#/.test(attr);
                    const isSplitArray = /^\$/.test(last);
                    const type = attr.replace(/^.+:(.+)$/, '$1');
                    const stringify = self._stringify[type] || self._stringify.auto;
                    const searchKey = attr.replace(/:.+$/, '').replace('$', '');
                    let value;
                    if (hasArray) {
                        const keys = attr.replace(/#(\w+).*$/, '$1');
                        const arr = lodash_1.default.get(data, keys);
                        lodash_1.default.forEach(arr, (d, j) => {
                            value = lodash_1.default.get(data, searchKey.replace(/#(\w+)/, `$1[${j}]`));
                            if (value !== undefined) {
                                if (isSplitArray) {
                                    value = lodash_1.default.map(value, (v) => {
                                        return stringify(v);
                                    }).join();
                                }
                                else {
                                    value = stringify(value);
                                }
                                cells.push({
                                    col: i + 1,
                                    row: row + j,
                                    value,
                                });
                                maxRow = Math.max(maxRow, row + j);
                            }
                        });
                    }
                    else {
                        value = lodash_1.default.get(data, searchKey);
                        if (value !== undefined) {
                            if (isSplitArray) {
                                value = lodash_1.default.map(value, (v) => {
                                    return stringify(v);
                                }).join();
                            }
                            else {
                                value = stringify(value);
                            }
                            cells.push({
                                col: i + 1,
                                row,
                                value,
                            });
                        }
                    }
                }
                catch (e) {
                    logging_1.default.logger.error('parse error.', attr, e.message);
                }
            });
        }
        lodash_1.default.forEach(worksheetData.datas, (data) => {
            if (opts.type === 'array') {
                const origin = lodash_1.default.transform(opts.ref_keys || this.opts.ref_keys, (result, refKey, i) => {
                    result[`__ref_${i}`] = data[refKey];
                }, {});
                const datas = lodash_1.default.get(data, opts.key);
                lodash_1.default.forEach(datas, (_data) => {
                    _data = lodash_1.default.assign({}, origin, _data);
                    addCell(_data);
                    maxRow += 1;
                    row = maxRow;
                });
            }
            else {
                addCell(data);
                maxRow += 1;
                row = maxRow;
            }
        });
        return {
            maxCol,
            maxRow,
            cells: lodash_1.default.sortBy(cells, ['row', 'col']),
            name: sheetName,
            opts,
        };
    }
    async updateWorksheetDatas(credentials, spreadsheetId, worksheetName, cells, callback) {
        if (arguments.length === 3) {
            cells = worksheetName;
            worksheetName = spreadsheetId;
            spreadsheetId = credentials;
        }
        else if (arguments.length === 4) {
            if (typeof cells === 'function') {
                callback = cells;
                cells = worksheetName;
                worksheetName = spreadsheetId;
                spreadsheetId = credentials;
                credentials = undefined;
            }
        }
        if (!callback) {
            callback = (err, result) => {
                if (err) {
                    throw err;
                }
                return result;
            };
        }
        const entry = [];
        lodash_1.default.forEach(cells, (cell) => {
            entry[cell.row] = entry[cell.row] || [];
            entry[cell.row][cell.col - 1] = cell.value;
        });
        try {
            const result = await api_1.default.batchCells(credentials, spreadsheetId, worksheetName, entry);
            return callback(null, result);
        }
        catch (e) {
            return callback(e);
        }
    }
    static get COLUMN_KEYMAP() {
        return COLUMN_KEYMAP;
    }
}
exports.default = new Spread2Json();
